"""
=========================================================

Custom Types (:mod:`rheedium.types`)

=========================================================

This submodule contains custom-defined types and PyTrees
for use throughout the rheedium package.
"""

from .crystal_types import *
from .custom_types import *
from .rheed_types import *



import jax
import jax.numpy as jnp
from beartype import beartype
from beartype.typing import Optional, Tuple
from jaxtyping import Array, Bool, Float, Int, jaxtyped

import rheedium as rh
from rheedium.types import *

jax.config.update("jax_enable_x64", True)


@jaxtyped(typechecker=beartype)
def incident_wavevector(
    lam_ang: Float[Array, ""], theta_deg: Float[Array, ""]
) -> Float[Array, "3"]:
    """
    Description
    -----------
    Build an incident wavevector k_in with magnitude (2π / λ),
    traveling mostly along +x, with a small angle theta from the x-y plane.

    Parameters
    ----------
    - `lam_ang` (Float[Array, ""]):
        Electron wavelength in angstroms
    - `theta_deg` (Float[Array, ""]):
        Grazing angle in degrees

    Returns
    -------
    - `k_in` (Float[Array, "3"]):
        The 3D incident wavevector (1/angstrom)
    """
    k_mag: Float[Array, ""] = 2.0 * jnp.pi / lam_ang
    theta: Float[Array, ""] = jnp.deg2rad(theta_deg)
    kx: Float[Array, ""] = k_mag * jnp.cos(theta)
    kz: Float[Array, ""] = -k_mag * jnp.sin(theta)
    k_in: Float[Array, "3"] = jnp.array([kx, 0.0, kz], dtype=jnp.float64)
    return k_in


@jaxtyped(typechecker=beartype)
def project_on_detector(
    k_out_set: Float[Array, "M 3"], detector_distance: Float[Array, ""]
) -> Float[Array, "M 2"]:
    """
    Description
    -----------
    Project wavevectors k_out onto a plane at x = detector_distance.
    Returns (M, 2) array of [Y, Z] coordinates on the detector.

    Parameters
    ----------
    - `k_out_set` (Float[Array, "M 3"]):
        (M, 3) array of outgoing wavevectors
    - `detector_distance` (Float[Array, ""):
        distance (in angstroms, or same unit) where screen is placed at x = L

    Returns
    -------
    - `coords` (Float[Array, "M 2"]):
        (M, 2) array of projected [Y, Z]
    """
    norms: Float[Array, "M 1"] = jnp.linalg.norm(k_out_set, axis=1, keepdims=True)
    directions: Float[Array, "M 3"] = k_out_set / (norms + 1e-12)
    t_vals: Float[Array, "M"] = detector_distance / (directions[:, 0] + 1e-12)
    Y: Float[Array, "M"] = directions[:, 1] * t_vals
    Z: Float[Array, "M"] = directions[:, 2] * t_vals
    coords: Float[Array, "M 2"] = jnp.stack([Y, Z], axis=-1)
    return coords


@jaxtyped(typechecker=beartype)
def find_kinematic_reflections(
    k_in: Float[Array, "3"],
    Gs: Float[Array, "M 3"],
    lam_ang: Float[Array, ""],
    z_sign: Optional[Float[Array, ""]] = jnp.asarray(1.0),
    tolerance: Optional[Float[Array, ""]] = jnp.asarray(0.05),
) -> Tuple[Int[Array, "K"], Float[Array, "K 3"]]:
    """
    Description
    -----------
    Returns indices of G for which ||k_in + G|| ~ 2π/lam
    and the z-component of (k_in + G) has the specified sign.

    Parameters
    ----------
    - `k_in` (Float[Array, "3"]):
        shape (3,)
    - `Gs` (Float[Array, "M 3]"):
        G vector
    - `lam_ang` (Float[Array, ""):
        electron wavelength in Å
    - `z_sign` (Float[Array, ""]):
        sign for z-component of k_out
    - `tolerance` (Float[Array, ""]):
        how close to the Ewald sphere in 1/Å

    Returns
    -------
    - `allowed_indices` (Int[Array, "K"]):
        Allowed indices that will kinematically reflect.
    - `k_out` (Float[Array, "K 3"]):
        Outgoing wavevectors (in 1/Å) for those reflections.
    """
    k_mag: Float[Array, ""] = 2.0 * jnp.pi / lam_ang
    k_out_candidates: Float[Array, "M 3"] = k_in[None, :] + Gs  # shape (M,3)
    norms: Float[Array, "M"] = jnp.linalg.norm(k_out_candidates, axis=1)
    cond_mag: Bool[Array, "M"] = jnp.abs(norms - k_mag) < tolerance
    cond_z: Bool[Array, "M"] = jnp.sign(k_out_candidates[:, 2]) == jnp.sign(z_sign)
    mask: Bool[Array, "M"] = jnp.logical_and(cond_mag, cond_z)
    allowed_indices: Int[Array, "K"] = jnp.where(mask)[0]
    k_out: Float[Array, "K 3"] = k_out_candidates[allowed_indices]
    return (allowed_indices, k_out)


def compute_kinematic_intensities(
    positions: Float[Array, "N 3"], G_allowed: Float[Array, "M 3"]
) -> Float[Array, "M"]:
    """
    Description
    -----------
    Given the atomic Cartesian positions (N,3) and the
    reciprocal vectors G_allowed (M,3),
    compute the kinematic intensity for each reflection:
        I(G) = | sum_j exp(i G·r_j) |^2
    ignoring atomic form factors, etc.

    Parameters
    ----------
    - `positions` (Float[Array, "N 3]):
        Atomic positions in Cartesian coordinates.
    - `G_allowed` (Float[Array, "M 3]):
        Reciprocal lattice vectors that satisfy reflection condition.

    Returns
    -------
    - `intensities` (Float[Array, "M"]):
        Intensities for each reflection.
    """

    def intensity_for_G(G_):
        phases = jnp.einsum("j,ij->i", G_, positions)
        re = jnp.sum(jnp.cos(phases))
        im = jnp.sum(jnp.sin(phases))
        return re * re + im * im

    intensities = jax.vmap(intensity_for_G)(G_allowed)
    return intensities


def simulate_rheed_pattern(
    crystal: CrystalStructure,
    voltage_kV: Optional[Float[Array, ""]] = jnp.asarray(10.0),
    theta_deg: Optional[Float[Array, ""]] = jnp.asarray(1.0),
    hmax: Optional[Int[Array, ""]] = jnp.asarray(3),
    kmax: Optional[Int[Array, ""]] = jnp.asarray(3),
    lmax: Optional[Int[Array, ""]] = jnp.asarray(1),
    tolerance: Optional[Float[Array, ""]] = jnp.asarray(0.05),
    detector_distance: Optional[Float[Array, ""]] = jnp.asarray(1000.0),
    z_sign: Optional[Float[Array, ""]] = jnp.asarray(1.0),
) -> RHEEDPattern:
    """
    Description
    -----------
    Compute a simple kinematic RHEED pattern for the given crystal.

    Parameters
    ----------
    - `crystal` (io.CrystalStructure):
        Crystal structure to simulate
    - `voltage_kV` (Float[Array, ""]):
        Accelerating voltage in kilovolts.
        Optional. Default: 10.0
    - `theta_deg` (Float[Array, ""]):
        Grazing angle in degrees
        Optional. Default: 1.0
    - `hmax, kmax, lmax` (Int[Array, ""]):
        Bounds on reciprocal lattice indices
        Optional. Default: 3, 3, 1
    - `tolerance` (Float[Array, ""]):
        How close to the Ewald sphere in 1/Å
        Optional. Default: 0.05
    - `detector_distance` (Float[Array, ""]):
        Distance from the sample to the detector plane in angstroms
        Optional. Default: 1000.0
    - `z_sign` (Float[Array, ""]):
        If +1, keep reflections with positive z in k_out
        Optional. Default: 1.0
    Returns
    -------
    - `pattern` (RHEEDPattern):
        A NamedTuple capturing reflection indices, k_out, and detector coords.

    Flow
    ----
    - Rebuild real-space cell vectors from (cell_lengths, cell_angles).
    - Generate reciprocal lattice points up to (hmax, kmax, lmax).
    - Compute electron wavelength in angstroms from voltage_kV (kV).
    - Build incident wavevector at grazing angle theta_deg.
    - Find which Gs satisfy the reflection condition: ||k_in + G|| ~ 2π/λ.
    - Project the resulting k_out onto a plane at x=detector_distance.
    - Return a RHEEDPattern with reflection info.
    """
    cell_vecs = rh.ucell.build_cell_vectors(
        crystal.cell_lengths[0],
        crystal.cell_lengths[1],
        crystal.cell_lengths[2],
        crystal.cell_angles[0],
        crystal.cell_angles[1],
        crystal.cell_angles[2],
    )
    Gs: Float[Array, "M 3"] = rh.ucell.generate_reciprocal_points(
        crystal=crystal,
        hmax=hmax,
        kmax=kmax,
        lmax=lmax,
        in_degrees=True,
    )
    lam_ang: Float[Array, ""] = rh.ucell.wavelength_ang(voltage_kV)
    k_in: Float[Array, "3"] = rh.simul.incident_wavevector(lam_ang, theta_deg)
    allowed_indices, k_out = rh.simul.find_kinematic_reflections(
        k_in=k_in, Gs=Gs, lam_ang=lam_ang, z_sign=z_sign, tolerance=tolerance
    )
    detector_points: Float[Array, "M 2"] = project_on_detector(
        k_out, jnp.asarray(detector_distance)
    )
    G_allowed = Gs[allowed_indices]
    atom_positions = crystal.cart_positions[:, :3]
    intensities: Float[Array, "M"] = rh.simul.compute_kinematic_intensities(
        positions=atom_positions, G_allowed=G_allowed
    )
    pattern = RHEEDPattern(
        G_indices=allowed_indices,
        k_out=k_out,
        detector_points=detector_points,
        intensities=intensities,
    )
    return pattern



import jax
import jax.numpy as jnp
from beartype import beartype
from beartype.typing import Optional, Tuple
from jaxtyping import Array, Bool, Float, Num, jaxtyped

import rheedium as rh
from rheedium.types import *

jax.config.update("jax_enable_x64", True)


@jaxtyped(typechecker=beartype)
def reciprocal_unitcell(unitcell: Num[Array, "3 3"]) -> Float[Array, "3 3"]:
    """
    Description
    -----------
    Calculate the reciprocal cell of a unit cell.

    Parameters
    ----------
    - `unitcell` (Num[Array, "3 3"]):
        The unit cell.

    Returns
    -------
    - `reciprocal_cell` (Float[Array, "3 3"]):
        The reciprocal cell.

    Flow
    ----
    - Calculate the reciprocal cell
    - Check if the matrix is well-conditioned
    - If not, replace the values with NaN
    """
    condition_number = jnp.linalg.cond(unitcell)
    is_well_conditioned = condition_number < 1e10
    reciprocal_cell_uncond: Float[Array, "3 3"] = (
        2 * jnp.pi * jnp.transpose(jnp.linalg.inv(unitcell))
    )
    reciprocal_cell: Float[Array, "3 3"] = jnp.where(
        is_well_conditioned,
        reciprocal_cell_uncond,
        jnp.full_like(reciprocal_cell_uncond, 0.0),
    )
    return reciprocal_cell


@jaxtyped(typechecker=beartype)
def reciprocal_uc_angles(
    unitcell_abc: Num[Array, "3"],
    unitcell_angles: Num[Array, "3"],
    in_degrees: Optional[bool] = True,
    out_degrees: Optional[bool] = False,
) -> Tuple[Float[Array, "3"], Float[Array, "3"]]:
    """
    Description
    -----------
    Calculate the reciprocal unit cell when the sides (a, b, c) and
    the angles (alpha, beta, gamma) are given.

    Parameters
    ----------
    - `unitcell_abc` (Num[Array, "3"]):
        The sides of the unit cell.
    - `unitcell_angles` (Num[Array, "3"]):
        The angles of the unit cell.
    - `in_degrees` (bool | None):
        Whether the angles are in degrees or radians.
        If None, it will be assumed that the angles are
        in degrees.
        Default is True.
    - `out_degrees` (bool | None):
        Whether the angles should be in degrees or radians.
        If None, it will be assumed that the angles should
        be in radians.
        Default is False.

    Returns
    -------
    - `reciprocal_abc` (Float[Array, "3"]):
        The sides of the reciprocal unit cell.
    - `reciprocal_angles` (Float[Array, "3"]):
        The angles of the reciprocal unit cell.

    Flow
    ----
    - Convert the angles to radians if they are in degrees
    - Calculate the cos and sin values of the angles
    - Calculate the volume factor of the unit cell
    - Calculate the unit cell volume
    - Calculate the reciprocal lattice parameters
    - Calculate the reciprocal angles
    - Convert the angles to degrees if they are in radians
    """
    if in_degrees:
        unitcell_angles = jnp.radians(unitcell_angles)
    cos_angles: Float[Array, "3"] = jnp.cos(unitcell_angles)
    sin_angles: Float[Array, "3"] = jnp.sin(unitcell_angles)
    volume_factor: Float[Array, ""] = jnp.sqrt(
        1 - jnp.sum(jnp.square(cos_angles)) + (2 * jnp.prod(cos_angles))
    )
    volume: Float[Array, ""] = jnp.prod(unitcell_abc) * volume_factor
    reciprocal_abc: Float[Array, "3"] = (
        jnp.array(
            [
                unitcell_abc[1] * unitcell_abc[2] * sin_angles[0],
                unitcell_abc[2] * unitcell_abc[0] * sin_angles[1],
                unitcell_abc[0] * unitcell_abc[1] * sin_angles[2],
            ]
        )
        / volume
    )
    reciprocal_angles = jnp.arccos(
        (cos_angles[:, None] * cos_angles[None, :] - cos_angles[None, :])
        / (sin_angles[:, None] * sin_angles[None, :])
    )
    reciprocal_angles: Float[Array, "3"] = jnp.array(
        [reciprocal_angles[1, 2], reciprocal_angles[2, 0], reciprocal_angles[0, 1]]
    )
    if out_degrees:
        reciprocal_angles = jnp.degrees(reciprocal_angles)
    return (reciprocal_abc, reciprocal_angles)


@jaxtyped(typechecker=beartype)
def get_unit_cell_matrix(
    unitcell_abc: Num[Array, "3"],
    unitcell_angles: Num[Array, "3"],
    in_degrees: Optional[bool] = True,
) -> Float[Array, "3 3"]:
    """
    Description
    -----------
    Calculate the transformation matrix for a unit cell using JAX.

    Parameters
    ----------
    - `unitcell_abc` (Num[Array, "3"]):
        Length of the unit cell edges (a, b, c) in Angstroms.
    - `unitcell_angles` (Num[Array, "3"]):
        Angles between the edges (alpha, beta, gamma) in degrees or radians.
    - `in_degrees` (bool | None):
        Whether the angles are in degrees or radians.
        Default is True.

    Returns
    -------
    - `matrix` (Float[Array, "3 3"]):
        3x3 transformation matrix

    Flow
    ----
    - Convert angles to radians if needed
    - Calculate trigonometric values
    - Calculate volume factor
    - Create the transformation matrix
    """
    angles_rad: Num[Array, "3"]
    if in_degrees:
        angles_rad = jnp.radians(unitcell_angles)
    else:
        angles_rad = unitcell_angles
    cos_angles: Float[Array, "3"] = jnp.cos(angles_rad)
    sin_angles: Float[Array, "3"] = jnp.sin(angles_rad)
    volume_factor: Float[Array, ""] = jnp.sqrt(
        1 - jnp.sum(jnp.square(cos_angles)) + (2 * jnp.prod(cos_angles))
    )
    matrix: Float[Array, "3 3"] = jnp.zeros(shape=(3, 3), dtype=jnp.float64)
    matrix = matrix.at[0, 0].set(unitcell_abc[0])
    matrix = matrix.at[0, 1].set(unitcell_abc[1] * cos_angles[2])
    matrix = matrix.at[0, 2].set(unitcell_abc[2] * cos_angles[1])
    matrix = matrix.at[1, 1].set(unitcell_abc[1] * sin_angles[2])
    matrix = matrix.at[1, 2].set(
        unitcell_abc[2]
        * (cos_angles[0] - cos_angles[1] * cos_angles[2])
        / sin_angles[2]
    )
    matrix = matrix.at[2, 2].set(unitcell_abc[2] * volume_factor / sin_angles[2])
    return matrix


@jaxtyped(typechecker=beartype)
def build_cell_vectors(
    a: scalar_num,
    b: scalar_num,
    c: scalar_num,
    alpha_deg: scalar_num,
    beta_deg: scalar_num,
    gamma_deg: scalar_num,
) -> Float[Array, "3 3"]:
    """
    Description
    -----------
    Convert (a, b, c, alpha, beta, gamma) into a 3x3 set of lattice vectors
    in Cartesian coordinates, using the standard crystallographic convention:

    - alpha = angle(b, c)
    - beta  = angle(a, c)
    - gamma = angle(a, b)

    Angles are in degrees.

    Parameters
    ----------
    - `a` (scalar_num):
        Length of the a-vector in Å
    - `b` (scalar_num):
        Length of the b-vector in Å
    - `c` (scalar_num):
        Length of the c-vector in Å
    - `alpha_deg` (scalar_num):
        Angle between b and c in degrees
    - `beta_deg` (scalar_num):
        Angle between a and c in degrees
    - `gamma_deg` (scalar_num):
        Angle between a and b in degrees

    Returns
    -------
    - `cell_vectors` (Float[Array, "3 3"]):
        The 3x3 array of lattice vectors in Cartesian coordinates.
        * cell_vectors[0] = a-vector
        * cell_vectors[1] = b-vector
        * cell_vectors[2] = c-vector

    Flow
    ----
    - Convert angles to radians
    - Calculate the a-vector along x
    - Calculate the b-vector in the x-y plane
    - Calculate the c-vector in full 3D
    - Stack the vectors to form the cell_vectors array
    """
    alpha: Float[Array, ""] = (alpha_deg * jnp.pi) / 180.0
    beta: Float[Array, ""] = (beta_deg * jnp.pi) / 180.0
    gamma: Float[Array, ""] = (gamma_deg * jnp.pi) / 180.0
    a_vec: Float[Array, "3"] = jnp.array([a, 0.0, 0.0])
    b_x: Float[Array, ""] = b * jnp.cos(gamma)
    b_y: Float[Array, ""] = b * jnp.sin(gamma)
    b_vec: Float[Array, "3"] = jnp.array([b_x, b_y, 0.0])
    c_x: Float[Array, ""] = c * jnp.cos(beta)
    c_y: Float[Array, ""] = c * (
        (jnp.cos(alpha) - jnp.cos(beta) * jnp.cos(gamma)) / jnp.sin(gamma)
    )
    c_z_sq: Float[Array, ""] = (c**2) - (c_x**2) - (c_y**2)
    c_z: Float[Array, ""] = jnp.sqrt(jnp.clip(c_z_sq, a_min=0.0))
    c_vec: Float[Array, "3"] = jnp.array([c_x, c_y, c_z])
    cell_vectors: Float[Array, "3 3"] = jnp.stack([a_vec, b_vec, c_vec], axis=0)
    return cell_vectors


@jaxtyped(typechecker=beartype)
def generate_reciprocal_points(
    crystal: CrystalStructure,
    hmax: Optional[scalar_int] = 3,
    kmax: Optional[scalar_int] = 3,
    lmax: Optional[scalar_int] = 1,
    in_degrees: Optional[bool] = True,
) -> Float[Array, "M 3"]:
    """
    Description
    -----------
    Generate a set of reciprocal-lattice vectors
    G_{hkl} = h a* + k b* + l c*
    for integer h, k, l in [-hmax..hmax], [-kmax..kmax], [-lmax..lmax].

    Utilizes `reciprocal_uc_angles` to find the reciprocal cell parameters
    from the direct (a,b,c,alpha,beta,gamma). Then constructs the
    (3x3) reciprocal-cell vectors via `build_cell_vectors`
    and forms the linear combinations.

    Parameters
    ----------
    - `crystal` (CrystalStructure)
        A NamedTuple containing cell_lengths and cell_angles (in degrees by default).
    - `hmax` (Optional[Int[Array, ""]]):
        Bounds on h. Default is 3.
    - `kmax` (Optional[Int[Array, ""]]):
        Bounds on k. Default is 3.
    - `lmax` (Optional[Int[Array, ""]]):
        Bounds on l. Default is 1.
    - `in_degrees` (Optional[bool]):
        If True, interpret the crystal.cell_angles as degrees.

    Returns
    -------
    - `Gs` (Float[Array, "M 3"]):
        The set of reciprocal-lattice vectors in inverse angstroms.
    """
    abc: Num[Array, "3"] = crystal.cell_lengths
    angles: Num[Array, "3"] = crystal.cell_angles
    rec_abc: Float[Array, "3"]
    rec_angles: Float[Array, "3"]
    rec_abc, rec_angles = rh.ucell.reciprocal_uc_angles(
        unitcell_abc=abc,
        unitcell_angles=angles,
        in_degrees=in_degrees,
        out_degrees=False,
    )
    rec_vectors: Float[Array, "3 3"] = rh.ucell.build_cell_vectors(
        rec_abc, rec_angles, in_degrees=False
    )
    a_star: Float[Array, "3"] = rec_vectors[0]
    b_star: Float[Array, "3"] = rec_vectors[1]
    c_star: Float[Array, "3"] = rec_vectors[2]
    hs: Num[Array, "n_h"] = jnp.arange(-hmax, hmax + 1)
    ks: Num[Array, "n_k"] = jnp.arange(-kmax, kmax + 1)
    ls: Num[Array, "n_l"] = jnp.arange(-lmax, lmax + 1)
    H: Num[Array, "n_h n_k n_l"]
    K: Num[Array, "n_h n_k n_l"]
    L: Num[Array, "n_h n_k n_l"]
    H, K, L = jnp.meshgrid(hs, ks, ls, indexing="ij")
    hkl: Num[Array, "M 3"] = jnp.stack([H.ravel(), K.ravel(), L.ravel()], axis=-1)

    def single_G(hkl_1d):
        h_ = hkl_1d[0]
        k_ = hkl_1d[1]
        l_ = hkl_1d[2]
        return (h_ * a_star) + (k_ * b_star) + (l_ * c_star)

    Gs: Float[Array, "M 3"] = jax.vmap(single_G)(hkl)
    return Gs


@jaxtyped(typechecker=beartype)
def atom_scraper(
    crystal: CrystalStructure,
    zone_axis: Num[Array, "3"],
    penetration_depth: Optional[scalar_num] = 0.0,
    eps: Optional[scalar_float] = 1e-3,
    max_atoms: Optional[scalar_int] = None,
) -> CrystalStructure:
    """
    Robustly filter atoms in `crystal` within `penetration_depth` from the top surface
    (along `zone_axis`). If `penetration_depth == 0.0`, only the topmost layer is returned.

    Parameters
    ----------
    - `crystal` (CrystalStructure):
        The input crystal structure.
    - `zone_axis` (Num[Array, "3"]):
        Surface normal axis in Cartesian space.
    - `penetration_depth` (scalar_num, optional):
        Thickness from the top layer to retain.
        Default: 0.0 (only top layer).
    - `eps` (scalar_float, optional):
        Numerical tolerance for identifying top layer atoms.
        Default: 1e-3.
    - `max_atoms` (scalar_int, optional):
        Maximum atoms to include. If None, defaults to the number of atoms in the input.

    Returns
    -------
    - `filtered_crystal` (CrystalStructure):
        New CrystalStructure with filtered atoms and adjusted cell.
    """
    orig_cell_vectors: Float[Array, "3 3"] = rh.ucell.build_cell_vectors(
        crystal.cell_lengths[0],
        crystal.cell_lengths[1],
        crystal.cell_lengths[2],
        crystal.cell_angles[0],
        crystal.cell_angles[1],
        crystal.cell_angles[2],
    )
    zone_axis_norm: Float[Array, ""] = jnp.linalg.norm(zone_axis)
    zone_axis_hat: Float[Array, "3"] = zone_axis / (zone_axis_norm + 1e-32)
    cart_xyz: Float[Array, "n 3"] = crystal.cart_positions[:, :3]
    dot_vals: Float[Array, "n"] = jnp.einsum("ij,j->i", cart_xyz, zone_axis_hat)
    d_max: Float[Array, ""] = jnp.max(dot_vals)
    dist_from_top: Float[Array, "n"] = d_max - dot_vals
    positive_distances = dist_from_top[dist_from_top > 1e-8]
    adaptive_eps = jnp.where(
        positive_distances.size > 0,
        jnp.maximum(eps, 2 * jnp.min(positive_distances)),
        eps,
    )
    is_top_layer_mode: Bool[Array, ""] = jnp.isclose(
        penetration_depth, jnp.asarray(0.0), atol=1e-8
    )
    mask: Bool[Array, "n"] = jnp.where(
        is_top_layer_mode,
        dist_from_top <= adaptive_eps,
        dist_from_top <= penetration_depth,
    )

    def gather_valid_positions(
        positions: Float[Array, "n 4"], gather_mask: Bool[Array, "n"]
    ) -> Float[Array, "m 4"]:
        return positions[gather_mask]

    filtered_frac: Float[Array, "m 4"] = gather_valid_positions(
        crystal.frac_positions, mask
    )
    filtered_cart: Float[Array, "m 4"] = gather_valid_positions(
        crystal.cart_positions, mask
    )
    original_height: Float[Array, ""] = jnp.max(dot_vals) - jnp.min(dot_vals)
    new_height: Float[Array, ""] = jnp.where(
        is_top_layer_mode, adaptive_eps, jnp.minimum(penetration_depth, original_height)
    )

    def scale_vector(
        vec: Float[Array, "3"],
        zone_axis_hat: Float[Array, "3"],
        old_height: Float[Array, ""],
        new_height: Float[Array, ""],
    ) -> Float[Array, "3"]:
        proj_mag: Float[Array, ""] = jnp.dot(vec, zone_axis_hat)
        parallel_comp: Float[Array, "3"] = proj_mag * zone_axis_hat
        perp_comp: Float[Array, "3"] = vec - parallel_comp
        scale_factor: Float[Array, ""] = jnp.where(
            old_height < 1e-32, 1.0, new_height / old_height
        )
        scaled_parallel: Float[Array, "3"] = scale_factor * parallel_comp
        return scaled_parallel + perp_comp

    def scale_if_needed(
        vec: Float[Array, "3"],
        zone_axis_hat: Float[Array, "3"],
        original_height: Float[Array, ""],
        new_height: Float[Array, ""],
    ) -> Float[Array, "3"]:
        needs_scaling: Bool[Array, ""] = jnp.abs(jnp.dot(vec, zone_axis_hat)) > 1e-8
        scaled: Float[Array, "3"] = scale_vector(
            vec, zone_axis_hat, original_height, new_height
        )
        return jnp.where(needs_scaling, scaled, vec)

    scaled_vectors = jnp.stack(
        [
            scale_if_needed(
                orig_cell_vectors[i], zone_axis_hat, original_height, new_height
            )
            for i in range(3)
        ]
    )
    new_lengths, new_angles = rh.ucell.compute_lengths_angles(scaled_vectors)
    filtered_crystal = CrystalStructure(
        frac_positions=filtered_frac,
        cart_positions=filtered_cart,
        cell_lengths=new_lengths,
        cell_angles=new_angles,
    )
    return filtered_crystal



from beartype.typing import NamedTuple
from jax.tree_util import register_pytree_node_class
from jaxtyping import Array, Float, Num

__all__ = ["CrystalStructure"]


@register_pytree_node_class
class CrystalStructure(NamedTuple):
    """
    Description
    -----------
    A JAX-compatible data structure representing a crystal structure with both
    fractional and Cartesian coordinates.

    Attributes
    ----------
    - `frac_positions` (Float[Array, "* 4"]):
        Array of shape (n_atoms, 4) containing atomic positions in fractional coordinates.
        Each row contains [x, y, z, atomic_number] where:
        - x, y, z: Fractional coordinates in the unit cell (range [0,1])
        - atomic_number: Integer atomic number (Z) of the element

    - `cart_positions` (Num[Array, "* 4"]):
        Array of shape (n_atoms, 4) containing atomic positions in Cartesian coordinates.
        Each row contains [x, y, z, atomic_number] where:
        - x, y, z: Cartesian coordinates in Ångstroms
        - atomic_number: Integer atomic number (Z) of the element

    - `cell_lengths` (Num[Array, "3"]):
        Unit cell lengths [a, b, c] in Ångstroms

    - `cell_angles` (Num[Array, "3"]):
        Unit cell angles [α, β, γ] in degrees.
        - α is the angle between b and c
        - β is the angle between a and c
        - γ is the angle between a and b

    Notes
    -----
    This class is registered as a PyTree node, making it compatible with JAX transformations
    like jit, grad, and vmap. The auxiliary data in tree_flatten is None as all relevant
    data is stored in JAX arrays.
    """

    frac_positions: Float[Array, "* 4"]
    cart_positions: Num[Array, "* 4"]
    cell_lengths: Num[Array, "3"]
    cell_angles: Num[Array, "3"]

    def tree_flatten(self):
        return (
            (
                self.frac_positions,
                self.cart_positions,
                self.cell_lengths,
                self.cell_angles,
            ),
            None,
        )

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        return cls(*children)



import jax.numpy as jnp
from beartype.typing import NamedTuple, Union
from jax.tree_util import register_pytree_node_class
from jaxtyping import Array, Float, Int

from rheedium.types import scalar_float, scalar_num

__all__ = ["RHEEDPattern", "RHEEDImage"]


@register_pytree_node_class
class RHEEDPattern(NamedTuple):
    """
    Description
    -----------
    A JAX-compatible data structure for representing RHEED patterns.

    Attributes
    ----------
    - `G_indices` (Int[Array, "*"]):
        Indices of reciprocal-lattice vectors that satisfy reflection
    - `k_out` (Float[Array, "M 3"]):
        Outgoing wavevectors (in 1/Å) for those reflections
    - `detector_points` (Float[Array, "M 2"]):
        (Y, Z) coordinates on the detector plane, in Ångstroms.
    - `intensities` (Float[Array, "M"]):
        Intensities for each reflection.
    """

    G_indices: Int[Array, "*"]
    k_out: Float[Array, "M 3"]
    detector_points: Float[Array, "M 2"]
    intensities: Float[Array, "M"]

    def tree_flatten(self):
        return (
            (self.G_indices, self.k_out, self.detector_points, self.intensities),
            None,
        )

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        return cls(*children)


@register_pytree_node_class
class RHEEDImage(NamedTuple):
    """
    Description
    -----------
    A PyTree for representing an experimental RHEED image.

    Attributes
    ----------
    - `img_array` (Float[Array, "H W"]):
        The image in 2D array format.
    - `incoming_angle` (scalar_float):
        The angle of the incoming electron beam in degrees.
    - `calibration` (Union[Float[Array, "2"], scalar_float]):
        Calibration factor for the image, either as a 2D array or a scalar.
        If scalar, then both the X and Y axes have the same calibration.
    - `electron_wavelength` (scalar_float):
        The wavelength of the electrons in Ångstroms.
    - `detector_distance` (scalar_float):
        The distance from the sample to the detector in Ångstroms.
    """

    img_array: Float[Array, "H W"]
    incoming_angle: scalar_float
    calibration: Union[Float[Array, "2"], scalar_float]
    electron_wavelength: scalar_float
    detector_distance: scalar_num

    def tree_flatten(self):
        return (
            (
                self.img_array,
                self.incoming_angle,
                self.calibration,
                self.electron_wavelength,
                self.detector_distance,
            ),
            None,
        )

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        return cls(*children)



"""
=========================================================

RHEEDIUM Package (:mod:`rheedium`)

=========================================================

This is the root of the rheedium package, containing submodules for:
- Data I/O (`inout`)
- Plotting (`plots`)
- Reconstructions (`recon`)
- Simulations (`simul`)
- Unit cell computations (`ucell`)
- Custom types (`types`)

Each submodule can be directly accessed after importing rheedium.
"""

from . import inout, plots, recon, simul, types, ucell



import jax
import matplotlib.pyplot as plt
import numpy as np
from beartype import beartype
from beartype.typing import List, Optional, Tuple
from matplotlib.colors import LinearSegmentedColormap
from scipy.interpolate import griddata

import rheedium as rh
from rheedium.types import *

jax.config.update("jax_enable_x64", True)


@beartype
def create_phosphor_colormap(
    name: Optional[str] = "phosphor",
) -> LinearSegmentedColormap:
    """
    Description
    -----------
    Create a custom colormap that simulates a phosphor screen appearance.
    The colormap transitions from black through a bright phosphorescent green,
    with a slight white bloom at maximum intensity.

    Parameters
    ----------
    - `name` (str, optional):
        Name for the colormap.
        Default is 'phosphor'

    Returns
    -------
    - `matplotlib.colors.LinearSegmentedColormap`
        Custom phosphor screen colormap
    """
    colors: List[
        Tuple[scalar_float, Tuple[scalar_float, scalar_float, scalar_float]]
    ] = [
        (0.0, (0.0, 0.0, 0.0)),
        (0.4, (0.0, 0.05, 0.0)),
        (0.7, (0.15, 0.85, 0.15)),
        (0.9, (0.45, 0.95, 0.45)),
        (1.0, (0.8, 1.0, 0.8)),
    ]
    positions: List[scalar_float] = [x[0] for x in colors]
    rgb_values: List[Tuple[scalar_float, scalar_float, scalar_float]] = [
        x[1] for x in colors
    ]
    red: List[Tuple[scalar_float, scalar_float, scalar_float]] = [
        (pos, rgb[0], rgb[0]) for pos, rgb in zip(positions, rgb_values)
    ]
    green: List[Tuple[scalar_float, scalar_float, scalar_float]] = [
        (pos, rgb[1], rgb[1]) for pos, rgb in zip(positions, rgb_values)
    ]
    blue: List[Tuple[scalar_float, scalar_float, scalar_float]] = [
        (pos, rgb[2], rgb[2]) for pos, rgb in zip(positions, rgb_values)
    ]
    cmap = LinearSegmentedColormap(name, {"red": red, "green": green, "blue": blue})
    return cmap


def plot_rheed(
    rheed_pattern: RHEEDPattern,
    grid_size: Optional[int] = 200,
    interp_type: Optional[str] = "cubic",
    cmap_name: Optional[str] = "phosphor",
) -> None:
    """
    Description
    -----------
    Interpolate the RHEED spots onto a uniform grid using various methods,
    then display using the phosphor colormap.

    The parameter `interp_type` controls which interpolation method is used:
      - "cubic"    => calls griddata(..., method="cubic")
      - "linear"   => calls griddata(..., method="linear")
      - "nearest"  => calls griddata(..., method="nearest")

    Parameters
    ----------
    - `rheed_pattern` (RHEEDPattern)
        Must have `detector_points` of shape (M, 2) and
        `intensities` of shape (M,).
    - `grid_size` (int)
        Controls how many grid points in Y and Z directions.
        Default is 200.
    - `interp_type` (str)
        Which interpolation approach to use. Default is "cubic".
        The parameter `interp_type` controls which interpolation method is used:
        - "cubic"    => calls griddata(..., method="cubic")
        - "linear"   => calls griddata(..., method="linear")
        - "nearest"  => calls griddata(..., method="nearest")
    - `cmap_name` (str):
        Name for your custom phosphor colormap.
        Default is 'phosphor'.
    """
    coords = rheed_pattern.detector_points
    Y = coords[:, 0]
    Z = coords[:, 1]
    intensities = rheed_pattern.intensities
    Y_np = np.asarray(Y)
    Z_np = np.asarray(Z)
    I_np = np.asarray(intensities)
    if interp_type in ("cubic", "linear", "nearest"):
        method = interp_type
    else:
        raise ValueError(
            "interp_type must be one of: 'cubic', 'linear', or 'nearest'. "
            f"Got: {interp_type}"
        )
    y_min, y_max = float(Y_np.min()), float(Y_np.max())
    z_min, z_max = float(Z_np.min()), float(Z_np.max())
    y_lin = np.linspace(y_min, y_max, grid_size)
    z_lin = np.linspace(z_min, z_max, grid_size)
    Yg, Zg = np.meshgrid(y_lin, z_lin, indexing="xy")
    grid_points = np.column_stack([Yg.ravel(), Zg.ravel()])
    interpolated = griddata(
        points=(Y_np, Z_np), values=I_np, xi=grid_points, method=method, fill_value=0.0
    )
    intensity_grid = interpolated.reshape((grid_size, grid_size))
    phosphor_cmap = rh.inout.create_phosphor_colormap(cmap_name)
    fig, ax = plt.subplots(figsize=(6, 6))
    cax = ax.imshow(
        intensity_grid.T,
        origin="lower",
        cmap=phosphor_cmap,
        extent=[y_min, y_max, z_min, z_max],
        aspect="equal",
        interpolation="bilinear",
    )
    cbar = fig.colorbar(cax, ax=ax)
    cbar.set_label("Interpolated Intensity (arb. units)")
    ax.set_title(f"RHEED Pattern ({method} interpolation)")
    ax.set_xlabel("Y (Å)")
    ax.set_ylabel("Z (Å)")
    plt.tight_layout()
    plt.show()



from beartype.typing import TypeAlias, Union
from jaxtyping import Array, Float, Integer, Num

__all__ = ["scalar_float", "scalar_int", "scalar_num", "non_jax_number"]

scalar_float: TypeAlias = Union[float, Float[Array, ""]]
scalar_int: TypeAlias = Union[int, Integer[Array, ""]]
scalar_num: TypeAlias = Union[int, float, Num[Array, ""]]
non_jax_number: TypeAlias = Union[int, float]



from pathlib import Path

import jax
import jax.numpy as jnp
from beartype import beartype
from beartype.typing import Optional, Tuple, Union
from jaxtyping import Array, Bool, Float, Real, jaxtyped

import rheedium as rh
from rheedium.types import *

jax.config.update("jax_enable_x64", True)


@jaxtyped(typechecker=beartype)
def wavelength_ang(voltage_kV: scalar_num) -> Float[Array, ""]:
    """
    Description
    -----------
    Calculates the relativistic electron wavelength
    in angstroms based on the microscope accelerating
    voltage.

    Because this is JAX - you assume that the input
    is clean, and you don't need to check for negative
    or NaN values. Your preprocessing steps should check
    for them - not the function itself.

    Parameters
    ----------
    - `voltage_kV` (scalar_num):
        The microscope accelerating voltage in kilo
        electronVolts

    Returns
    -------
    - `in_angstroms (Float[Array, ""]):
        The electron wavelength in angstroms

    Flow
    ----
    - Calculate the electron wavelength in meters
    - Convert the wavelength to angstroms
    """
    m: Float[Array, ""] = jnp.float64(9.109383e-31)
    e: Float[Array, ""] = jnp.float64(1.602177e-19)
    c: Float[Array, ""] = jnp.float64(299792458.0)
    h: Float[Array, ""] = jnp.float64(6.62607e-34)
    voltage: Float[Array, ""] = jnp.multiply(jnp.float64(voltage_kV), jnp.float64(1000))
    eV = jnp.multiply(e, voltage)
    numerator: Float[Array, ""] = jnp.multiply(jnp.square(h), jnp.square(c))
    denominator: Float[Array, ""] = jnp.multiply(eV, ((2 * m * jnp.square(c)) + eV))
    wavelength_meters: Float[Array, ""] = jnp.sqrt(numerator / denominator)
    in_angstroms: Float[Array, ""] = 1e10 * wavelength_meters
    return in_angstroms


@jaxtyped(typechecker=beartype)
def angle_in_degrees(u: Float[Array, "c"], v: Float[Array, "c"]) -> Float[Array, ""]:
    """
    Description
    -----------
    Calculate the angle in degrees between two vectors u and v.

    Parameters
    ----------
    - `u` (Float[Array, "c"]):
        The first vector.
    - `v` (Float[Array, "c"]):
        The second vector.

    Returns
    -------
    - `cos_val_degrees` (Float[Array, ""]):
        The angle in degrees between u and v.

    Flow
    ----
    - Calculate the dot product of u and v
    - Calculate the cosine of the angle
    - Convert the cosine to degrees
    """
    dot_uv: Float[Array, ""] = jnp.dot(u, v)
    cos_val: Float[Array, ""] = dot_uv / (
        jnp.linalg.norm(u) * jnp.linalg.norm(v) + 1e-32
    )
    cos_val_clamped: Float[Array, ""] = jnp.clip(cos_val, -1.0, 1.0)
    cos_val_degrees: Float[Array, ""] = (jnp.arccos(cos_val_clamped) * 180.0) / jnp.pi
    return cos_val_degrees


@jaxtyped(typechecker=beartype)
def compute_lengths_angles(
    cell_vectors: Float[Array, "3 3"],
) -> Tuple[Float[Array, "3"], Float[Array, "3"]]:
    """
    Description
    -----------
    Given a (3, 3) array of lattice vectors (a_vec, b_vec, c_vec),
    compute (a, b, c) in Å and (alpha, beta, gamma) in degrees.

    Parameters
    ----------
    - `cell_vectors` (Float[Array, "3 3"]):
        The 3x3 array of lattice vectors in Cartesian coordinates.

    Returns
    -------
    - `lengths` (Float[Array, "3"]):
        The lengths of the lattice vectors (a, b, c) in Å.
    - `angles` (Float[Array, "3"]):
        The angles between the lattice vectors (alpha, beta, gamma) in degrees.
    """
    a_vec: Float[Array, "3"] = cell_vectors[0]
    b_vec: Float[Array, "3"] = cell_vectors[1]
    c_vec: Float[Array, "3"] = cell_vectors[2]
    a_len: Float[Array, ""] = jnp.linalg.norm(a_vec)
    b_len: Float[Array, ""] = jnp.linalg.norm(b_vec)
    c_len: Float[Array, ""] = jnp.linalg.norm(c_vec)
    alpha: Float[Array, ""] = rh.ucell.angle_in_degrees(b_vec, c_vec)
    beta: Float[Array, ""] = rh.ucell.angle_in_degrees(a_vec, c_vec)
    gamma: Float[Array, ""] = rh.ucell.angle_in_degrees(a_vec, b_vec)
    lengths: Float[Array, "3"] = jnp.array([a_len, b_len, c_len])
    angles: Float[Array, "3"] = jnp.array([alpha, beta, gamma])
    return (lengths, angles)


@jaxtyped(typechecker=beartype)
def parse_cif_and_scrape(
    cif_path: Union[str, Path],
    zone_axis: Real[Array, "3"],
    thickness_xyz: Real[Array, "3"],
    tolerance: Optional[scalar_float] = 1e-3,
) -> CrystalStructure:
    """
    Description
    -----------
    Parse a CIF file, apply symmetry operations to obtain all equivalent
    atomic positions, and scrape (filter) atoms within specified thickness
    along a given zone axis.

    Parameters
    ----------
    - `cif_path` (Union[str, Path]):
        Path to the CIF file.
    - `zone_axis` (Real[Array, "3"]):
        Vector indicating the zone axis direction (surface normal) in
        Cartesian coordinates.
    - `thickness_xyz` (Real[Array, "3"]):
        Thickness along x, y, z directions in Ångstroms; currently,
        only thickness_xyz[2] (z-direction)
        is used to filter atoms along the provided zone axis.
    - `tolerance` (scalar_float, optional):
        Numerical tolerance parameter reserved for future use.
        Default is 1e-3.

    Returns
    -------
    - `filtered_crystal` (CrystalStructure):
        Crystal structure containing atoms filtered within the specified thickness.

    Notes
    -----
    - The provided `zone_axis` is normalized internally.
    - Current implementation uses thickness only along the zone axis
        direction (z-component of `thickness_xyz`).
    - The `tolerance` parameter is reserved for compatibility and future
        functionality.
    """
    crystal: CrystalStructure = rh.inout.parse_cif(cif_path)
    cart_positions: Float[Array, "n 3"] = crystal.cart_positions[:, :3]
    atomic_numbers: Float[Array, "n"] = crystal.cart_positions[:, 3]
    zone_axis_norm: Float[Array, ""] = jnp.linalg.norm(zone_axis)
    zone_axis_hat: Float[Array, "3"] = zone_axis / (zone_axis_norm + 1e-12)
    projections: Float[Array, "n"] = cart_positions @ zone_axis_hat
    min_proj: Float[Array, ""] = jnp.min(projections)
    max_proj: Float[Array, ""] = jnp.max(projections)
    center_proj: Float[Array, ""] = (max_proj + min_proj) / 2.0
    half_thickness: Float[Array, ""] = thickness_xyz[2] / 2.0
    mask: Bool[Array, "n"] = jnp.abs(projections - center_proj) <= half_thickness
    filtered_cart_positions: Float[Array, "m 3"] = cart_positions[mask]
    filtered_atomic_numbers: Float[Array, "m"] = atomic_numbers[mask]
    cell_vectors: Float[Array, "3 3"] = rh.ucell.build_cell_vectors(
        crystal.cell_lengths[0],
        crystal.cell_lengths[1],
        crystal.cell_lengths[2],
        crystal.cell_angles[0],
        crystal.cell_angles[1],
        crystal.cell_angles[2],
    )
    cell_inv: Float[Array, "3 3"] = jnp.linalg.inv(cell_vectors)
    filtered_frac_positions: Float[Array, "m 3"] = (
        filtered_cart_positions @ cell_inv
    ) % 1.0
    filtered_crystal: CrystalStructure = CrystalStructure(
        frac_positions=jnp.column_stack(
            [filtered_frac_positions, filtered_atomic_numbers]
        ),
        cart_positions=jnp.column_stack(
            [filtered_cart_positions, filtered_atomic_numbers]
        ),
        cell_lengths=crystal.cell_lengths,
        cell_angles=crystal.cell_angles,
    )
    return filtered_crystal



"""
=========================================================

Unit Cell Operations (:mod:`rheedium.ucell`)

=========================================================

This package contains the modules for the calculations of
unit cell operations and conversion to Ewald sphere.
"""

from .helper import *
from .unitcell import *



import fractions
import json
import re
from pathlib import Path

import jax
import jax.numpy as jnp
from beartype import beartype
from beartype.typing import List, Union
from jaxtyping import Array, Float, Num, jaxtyped

import rheedium as rh
from rheedium.types import *

DEFAULT_ATOMIC_NUMBERS_PATH = (
    Path(__file__).resolve().parents[3] / "data" / "atomic_numbers.json"
)


@beartype
def load_atomic_numbers(path: str = str(DEFAULT_ATOMIC_NUMBERS_PATH)) -> dict[str, int]:
    """
    Description
    -----------
    Load the atomic numbers mapping from a JSON file.

    Parameters
    ----------
    - `path` (str, optional):
        Path to the atomic numbers JSON file.
        Defaults to '<project_root>/data/atomic_numbers.json'.

    Returns
    -------
    - `atomic_numbers` (dict[str, int]):
        Dictionary mapping element symbols to atomic numbers.
    """
    with open(path, "r") as f:
        atomic_numbers = json.load(f)
    return atomic_numbers


@jaxtyped(typechecker=beartype)
def parse_cif(cif_path: Union[str, Path]) -> CrystalStructure:
    """
    Description
    -----------
    Parse a CIF file into a JAX-compatible CrystalStructure.

    Parameters
    ----------
    - `cif_path` (Union[str, Path]):
        Path to the CIF file.

    Returns
    -------
    `CrystalStructure`:
        Parsed crystal structure object with fractional and Cartesian coordinates.

        Attributes:

        - `frac_positions` (Float[Array, "* 4"]):
            Array of shape (n_atoms, 4) containing atomic positions in fractional coordinates.
            Each row contains [x, y, z, atomic_number] where:
            - x, y, z: Fractional coordinates in the unit cell (range [0,1])
            - atomic_number: Integer atomic number (Z) of the element

        - `cart_positions` (Num[Array, "* 4"]):
            Array of shape (n_atoms, 4) containing atomic positions in Cartesian coordinates.
            Each row contains [x, y, z, atomic_number] where:
            - x, y, z: Cartesian coordinates in Ångstroms
            - atomic_number: Integer atomic number (Z) of the element

        - `cell_lengths` (Num[Array, "3"]):
            Unit cell lengths [a, b, c] in Ångstroms

        - `cell_angles` (Num[Array, "3"]):
            Unit cell angles [α, β, γ] in degrees.
            - α is the angle between b and c
            - β is the angle between a and c
            - γ is the angle between a and b
    """
    cif_path = Path(cif_path)
    if not cif_path.exists():
        raise FileNotFoundError(f"CIF file not found: {cif_path}")
    if cif_path.suffix.lower() != ".cif":
        raise ValueError(f"File must have .cif extension: {cif_path}")
    cif_text = cif_path.read_text()
    atomic_numbers = load_atomic_numbers()

    def extract_param(name: str) -> float:
        match = re.search(rf"{name}\s+([0-9.]+)", cif_text)
        if match:
            return float(match.group(1))
        else:
            raise ValueError(f"Failed to parse {name} from CIF.")

    a = extract_param("_cell_length_a")
    b = extract_param("_cell_length_b")
    c = extract_param("_cell_length_c")
    alpha = extract_param("_cell_angle_alpha")
    beta = extract_param("_cell_angle_beta")
    gamma = extract_param("_cell_angle_gamma")
    cell_lengths: Num[Array, "3"] = jnp.array([a, b, c], dtype=jnp.float64)
    cell_angles: Num[Array, "3"] = jnp.array([alpha, beta, gamma], dtype=jnp.float64)
    lines = cif_text.splitlines()
    atom_site_columns = []
    positions_list = []
    in_atom_site_loop = False
    for line in lines:
        stripped_line = line.strip()
        if stripped_line.lower().startswith("loop_"):
            in_atom_site_loop = False
            atom_site_columns = []
            continue
        if stripped_line.startswith("_atom_site_"):
            atom_site_columns.append(stripped_line)
            in_atom_site_loop = True
            continue
        if in_atom_site_loop and stripped_line and not stripped_line.startswith("_"):
            tokens = stripped_line.split()
            if len(tokens) != len(atom_site_columns):
                continue
            required_cols = [
                "_atom_site_type_symbol",
                "_atom_site_fract_x",
                "_atom_site_fract_y",
                "_atom_site_fract_z",
            ]
            if not all(col in atom_site_columns for col in required_cols):
                continue
            col_indices = {col: atom_site_columns.index(col) for col in required_cols}
            element_symbol = tokens[col_indices["_atom_site_type_symbol"]]
            frac_x = float(tokens[col_indices["_atom_site_fract_x"]])
            frac_y = float(tokens[col_indices["_atom_site_fract_y"]])
            frac_z = float(tokens[col_indices["_atom_site_fract_z"]])
            atomic_number = atomic_numbers.get(element_symbol)
            if atomic_number is None:
                raise ValueError(f"Unknown element symbol: {element_symbol}")
            positions_list.append([frac_x, frac_y, frac_z, atomic_number])
    if not positions_list:
        raise ValueError("No atomic positions found in CIF.")
    frac_positions: Float[Array, "* 4"] = jnp.array(positions_list, dtype=jnp.float64)
    cell_vectors: Float[Array, "3 3"] = rh.ucell.build_cell_vectors(
        a, b, c, alpha, beta, gamma
    )
    cart_coords: Float[Array, "* 3"] = frac_positions[:, :3] @ cell_vectors
    cart_positions: Float[Array, "* 4"] = jnp.column_stack(
        (cart_coords, frac_positions[:, 3])
    )
    sym_ops = []
    lines = cif_text.splitlines()
    collect_sym_ops = False
    for line in lines:
        stripped_line = line.strip()
        if stripped_line.startswith("_symmetry_equiv_pos_as_xyz"):
            collect_sym_ops = True
            continue
        if collect_sym_ops:
            if stripped_line.startswith("'") and stripped_line.endswith("'"):
                op_clean = stripped_line.strip("'").strip()
                sym_ops.append(op_clean)
            elif stripped_line.startswith('"') and stripped_line.endswith('"'):
                op_clean = stripped_line.strip('"').strip()
                sym_ops.append(op_clean)
            else:
                if sym_ops:
                    break

    if not sym_ops:
        sym_ops = ["x,y,z"]

    crystal = CrystalStructure(
        frac_positions=frac_positions,
        cart_positions=cart_positions,
        cell_lengths=cell_lengths,
        cell_angles=cell_angles,
    )
    expanded_crystal = symmetry_expansion(crystal, sym_ops, tolerance=1.0)
    return expanded_crystal


@jaxtyped(typechecker=beartype)
def symmetry_expansion(
    crystal: CrystalStructure,
    sym_ops: List[str],
    tolerance: scalar_float = 1.0,
) -> CrystalStructure:
    """
    Description
    -----------
    Apply symmetry operations to expand fractional positions and remove duplicates.

    Parameters
    ----------
    - `crystal` (CrystalStructure):
        The initial crystal structure with symmetry-independent positions.
    - `sym_ops` (List[str]):
        List of symmetry operations as strings from the CIF file.
        Example: ["x,y,z", "-x,-y,z", ...]
    - `tolerance` (scalar_float):
        Distance tolerance in angstroms for duplicate atom removal.
        Default: 1.0 Å.

    Returns
    -------
    - `expanded_crystal` (CrystalStructure):
        Symmetry-expanded crystal structure without duplicates.
    """
    frac_positions = crystal.frac_positions
    expanded_positions = []

    def parse_sym_op(op_str: str):
        def op(pos):
            replacements = {"x": pos[0], "y": pos[1], "z": pos[2]}
            components = op_str.lower().replace(" ", "").split(",")

            def eval_comp(comp):
                comp = comp.replace("-", "+-")
                terms = comp.split("+")
                total = 0.0
                for term in terms:
                    if not term:
                        continue
                    coeff = 1.0
                    for var in ("x", "y", "z"):
                        if var in term:
                            part = term.split(var)[0]
                            coeff = float(fractions.Fraction(part)) if part else 1.0
                            total += coeff * replacements[var]
                            break
                    else:
                        total += float(fractions.Fraction(term))
                return total

            return jnp.array([eval_comp(c) for c in components])

        return op

    ops = [parse_sym_op(op) for op in sym_ops]
    for pos in frac_positions:
        xyz, atomic_number = pos[:3], pos[3]
        for op in ops:
            new_xyz = jnp.mod(op(xyz), 1.0)
            expanded_positions.append(jnp.concatenate([new_xyz, atomic_number[None]]))
    expanded_positions = jnp.array(expanded_positions)
    cart_positions = expanded_positions[:, :3] @ rh.ucell.build_cell_vectors(
        *crystal.cell_lengths, *crystal.cell_angles
    )

    def deduplicate(positions, tol):
        def dist(p1, p2):
            return jnp.sqrt(jnp.sum((p1 - p2) ** 2))

        def unique_cond(carry, pos):
            unique, count = carry
            diff = unique - pos
            dist_sq = jnp.sum(diff**2, axis=1)
            is_dup = jnp.any(dist_sq < tol**2)
            unique = jax.lax.cond(
                is_dup,
                lambda u: u,
                lambda u: u.at[count].set(pos),
                unique,
            )
            count += jnp.logical_not(is_dup)
            return (unique, count), None

        unique_init = jnp.zeros_like(positions)
        unique_init = unique_init.at[0].set(positions[0])
        count_init = 1
        (unique_final, final_count), _ = jax.lax.scan(
            unique_cond, (unique_init, count_init), positions[1:]
        )
        return unique_final[:final_count]

    unique_cart = deduplicate(cart_positions, tolerance)
    cell_inv = jnp.linalg.inv(
        rh.ucell.build_cell_vectors(*crystal.cell_lengths, *crystal.cell_angles)
    )
    unique_frac = (unique_cart @ cell_inv) % 1.0
    atomic_numbers = expanded_positions[:, 3][: unique_cart.shape[0]]
    expanded_crystal = CrystalStructure(
        frac_positions=jnp.column_stack([unique_frac, atomic_numbers]),
        cart_positions=jnp.column_stack([unique_cart, atomic_numbers]),
        cell_lengths=crystal.cell_lengths,
        cell_angles=crystal.cell_angles,
    )
    return expanded_crystal
